; SCALE
; 1 Patch = 5 metres
; 10 ticks = 1 second

globals [
  lanes                           ; A list of y coordinates of different lanes
  jam-distance                    ; Minimum distance to be maintained (IDM)
  acceleration-exponent           ; Extent by which acceleration decreases while approaching desired velocity (IDM)
  threshold                       ; Acceleration Threshold (MOBIL)
  crash-count                     ; Total crashes that occurs during the simulation
  any-lane-blocked?               ; Toggles blockade
  blockade-xcor                   ; x-cor of blockade
  red-zone-start-xcor             ; Starting x-cor of red-zone
  red-zone-end-xcor               ; Ending x-cor of red-zone
  throughput                      ; Measures total cars that pass through the end of the world per tick
  cruising-speed                  ; Maximum speed of cars
  total-travel-time               ; Cumulative time taken by every car to pass through the world
  spawn-count                     ; Total number of cars spawned
  non-realistic-brake-count       ; A guard to check cars exceed deceleration limit
  total-time-halted               ; Cumulative time that cars remain halted for
  cumulative-cars-in-world-count  ; Cumulative count of cars in the world per tick
]

; Breeds
breed [cars car]
breed [blockades blockade]

turtles-own [
  speed                        ; (float) The current speed of the vehicle
  preferred-speed              ; (float) The maximum speed of the vehicle (different for all drivers)
  reaction-time                ; (float) The time required to react to an event
  target-lane                  ; (int)   The driver's desired lane
  max-acceleration             ; (float) The vehicle's max-acceleration
  comfortable-deceleration     ; (float) The driver's comfortable-deceleration
  max-deceleration             ; (float) The driver's max-deceleration
  crashed?                     ; (bool)  If the driver has crashed the vehicle
  crash-tick                   ; (int)   The time of crash of vehicle
  politeness-factor            ; (float) Measure of politeness (MOBIL)
  curr-acc                     ; (float) Current acceleration (MOBIL)
  blockade?                    ; (bool)  True only if the turtle is a blockade
  entry-tick                   ; (int)   The time at which the turtle is spawned
  slow-down-reqs               ; (int)   The number of times the car has been asked to slow down (STR1)
  who-to-allow                 ; (id)    The id of the car that should be allowed to merge in front
]


; ============= DRAW-COMMANDS =============

to draw-road
  ask patches [ set pcolor green - random-float 0.5 ]                                             ; The road is surrounded by green grass of varying shades
  set lanes n-values number-of-lanes [ n -> number-of-lanes - (n * 2) - 1 ]
  ask patches with [ abs pycor <= number-of-lanes ] [ set pcolor grey - 2.5 + random-float 0.25 ] ; The road itself is varying shades of grey
  draw-road-lines
end

to draw-road-lines
  let y (last lanes) - 1                ; Start below the "lowest" lane
  while [ y <= first lanes + 1 ] [
    if not member? y lanes [
      ifelse abs y = number-of-lanes    ; Draw lines on road patches that are not part of a lane
        [ draw-line y yellow 0 ]        ; Yellow for the sides of the road
        [ draw-line y white 0.5 ]       ; Dashed white between lanes
    ]
    set y y + 1                         ; Move up one patch
  ]
end

to draw-line [ y line-color gap ]
  ; We use a temporary turtle to draw the line:
  ; - with a gap of zero, we get a continuous line;
  ; - with a gap greater than zero, we get a dashed line.
  create-turtles 1 [
    setxy (min-pxcor - 0.5) y
    hide-turtle
    set color line-color
    set heading 90
    repeat world-width [
      pen-up
      forward gap
      pen-down
      forward (1 - gap)
    ]
    die
  ]
end

; ============= END DRAW-COMMANDS =============


; ============= SPAWN-COMMANDS =============

; Spawn a car in the given lane (ycord) with a speed determined by IDM (spawn-speed)
to spawn-car [ ycord ]
  set spawn-count spawn-count + 1
  let future-lead-vehicle min-one-of (other turtles with [ycor = ycord and not blockade?] ) [ xcor ]
  let curr-speed 0.625

  create-turtles (1)[
    set breed cars
    set shape "cars-top"
    set color vehicle-color
    move-to patch ((- world-width + 1) / 2) ycord
    set target-lane ycor
    set heading 90
    set preferred-speed 0.625
    set max-acceleration 0.001939
    set comfortable-deceleration 0.00518
    set max-deceleration 0.00794
    set slow-down-reqs 0
    set crashed? false
    set jam-distance 0.4
    set reaction-time 16
    set blockade? false
    set politeness-factor random-normal politeness-factor-mean politeness-factor-deviation
    set entry-tick ticks
    set who-to-allow nobody
    set speed curr-speed
    set speed ifelse-value (future-lead-vehicle != nobody)[ precision(spawn-speed future-lead-vehicle) 6 ][ curr-speed ]
  ]
end


; Calculate comfortable speed based on IDM, given a lead vehicle
to-report spawn-speed [ leading-vehicle ]
  let constant 2 * sqrt( max-acceleration * comfortable-deceleration )
  report (( [ speed ] of leading-vehicle ) - ( constant * reaction-time ) + ( sqrt( (( ( constant * reaction-time ) - [ speed ] of leading-vehicle ) ^ 2) - (4 * constant * ( jam-distance - (precision (([ xcor ] of leading-vehicle - xcor)) 6) )) ) ) ) / 2
end


; Routine that randomly spawns cars when there is enough space
to spawn-cars
  foreach lanes [ lane ->

    ; Make sure there are no turtles in the first column of the world.
    if not any? turtles with [ xcor <= ((- world-width + 5) / 2) and ycor = lane ][
      ; Randomize spawn
      let spawn? random 2
      if spawn? = 1[
        spawn-car lane
      ]
    ]

  ]
end


; Place a blockade in the given lane (ycord)
to spawn-blockade [ ycord ]
  create-blockades (1)[
    set breed blockades
    set shape "cars-top"
    set color black
    move-to patch blockade-xcor ycord
    set target-lane ycor
    set heading 90
    set blockade? true
  ]
end

; ============= END SPAWN-COMMANDS =============


; ============= IDM =============

to-report comfortable-gap [ vehicle-ahead ]
  ifelse (vehicle-ahead = nobody)[ report precision ( jam-distance + ( reaction-time * speed )) 6]
  [ report precision ( jam-distance + ( reaction-time * speed ) + (( speed * max( list 0 ( speed - [ speed ] of vehicle-ahead ))/( 2 * sqrt( max-acceleration * comfortable-deceleration ))))) 6]
end

to-report comfortable-speed [ vehicle-ahead gap ]
  report min(list (cruising-speed) (precision ( (gap - jam-distance) / (max( list 0 ( speed - [ speed ] of vehicle-ahead ))/( 2 * sqrt( max-acceleration * comfortable-deceleration )) + reaction-time )) 6))
end

to-report acceleration
  let out 0
  let vehicle-ahead min-one-of ( other turtles in-cone world-width 180 with [ y-distance <= 1 and x-distance >= 0 ] ) [ distance myself ]
  ifelse (vehicle-ahead = nobody)[
    if blockade? [report 0]
    set curr-acc precision ( max-acceleration * ( 1 - (( speed / preferred-speed ) ^ acceleration-exponent ))) 6
    set out precision ( max-acceleration * ( 1 - (( speed / preferred-speed ) ^ acceleration-exponent ))) 6
  ][
    if blockade? [report 0]
    set curr-acc precision ( max-acceleration * ( 1 - (( speed / preferred-speed ) ^ acceleration-exponent ) - (( comfortable-gap vehicle-ahead / leading-gap vehicle-ahead ) ^ 2 ))) 6
    set out curr-acc
  ]
  report max(list (-1 * max-deceleration) out)
end

to-report leading-gap [ vehicle-ahead ]
  let gap precision ([ xcor ] of vehicle-ahead - xcor - 1) 6
  report ifelse-value (gap = 0)[ 0.001 ][gap] ; Avoid division by zero (when called to calculate future possibilities). Lane change is blocked even if favourable by move-to-target-lane.
end

to calculate-speeds
  if acceleration < -1 * max-deceleration [ set non-realistic-brake-count non-realistic-brake-count + 1 ]
  set speed max(list 0 precision (speed + acceleration) 6 )
end

to-report calculate-realistic-speed
  let acc acceleration
  set acc max(list (-1 * max-deceleration) acc)
  report max(list 0 precision (speed + acc) 6 )
end

; ============= END-IDM =============


; ============= MOBIL =============

to-report future-acceleration [ leading-vehicle ]
  if blockade? [report 0]
  ifelse (leading-vehicle = nobody)[ report precision ( max-acceleration * ( 1 - (( speed / preferred-speed ) ^ acceleration-exponent ))) 6 ]
  [
    report precision ( max-acceleration * ( 1 - (( speed / preferred-speed ) ^ acceleration-exponent ) - (( comfortable-gap leading-vehicle / leading-gap leading-vehicle) ^ 2 ))) 6
  ]
end

to-report lane-change-utility [ lane ]
  set heading -90
  let future-following-vehicle min-one-of ( other turtles in-cone world-width 150 with [ abs( ycor - lane ) <= 0.75 ] ) [ distance myself ]
  let current-following-vehicle min-one-of ( other turtles in-cone world-width 150 with [ y-distance <= 0.75 ] ) [ distance myself ]
  set heading 90
  let future-leading-vehicle min-one-of ( other turtles in-cone world-width 150 with [ abs( ycor - lane ) <= 0.75 ] ) [ distance myself ]
  let current-leading-vehicle min-one-of ( other turtles in-cone world-width 150 with [ y-distance <= 0.75 ] ) [ distance myself ]
  let utility future-acceleration future-leading-vehicle - acceleration
  if (future-following-vehicle != nobody) [ set utility utility + politeness-factor * ( [ future-acceleration myself ] of future-following-vehicle - [ acceleration ] of future-following-vehicle ) ]
  if (current-following-vehicle != nobody) [ set utility utility + politeness-factor * ( [ future-acceleration current-leading-vehicle ] of current-following-vehicle - [ acceleration ] of current-following-vehicle ) ]
  report utility
end

to check-lane-change
  let other-lanes remove ycor lanes
  let currbest threshold
  if not empty? other-lanes [
    let min-dist min map [ y -> abs (y - ycor) ] other-lanes
    let closest-lanes filter [ y -> abs (y - ycor) = min-dist ] other-lanes
    foreach closest-lanes[ [ lane ] ->
      set heading -90
      let future-follower min-one-of ( other turtles with [ abs( ycor - lane ) <= 0.75 and xcor < [xcor] of myself ] ) [ distance myself ]
      set heading 90
      if (future-follower = nobody) or (abs([ future-acceleration myself ] of future-follower) < [ abs(max-deceleration) ] of future-follower)[
        if not any? other turtles with [ ycor = lane and abs(xcor - [xcor] of myself) < 1 ] and lane-change-utility lane > currbest [
          set currbest lane-change-utility lane
          set target-lane lane
        ]
      ]
    ]
  ]
end

; ============= END-MOBIL =============


; ============= MOBIL - modified utility =============

to-report lane-change-utility-force [ lane ]
  let block one-of blockades
  ifelse block != nobody and lane = [ycor] of block [
    report -1
  ][
    report lane-change-utility lane
  ]

end

to check-lane-change-force
  let other-lanes remove ycor lanes
  let currbest threshold
  if not empty? other-lanes [
    let min-dist min map [ y -> abs (y - ycor) ] other-lanes
    let closest-lanes filter [ y -> abs (y - ycor) = min-dist ] other-lanes
    foreach closest-lanes[ [ lane ] ->
      set heading -90
      let future-follower min-one-of ( other turtles in-cone world-width 150 with [ abs( ycor - lane ) <= 0.75 ] ) [ distance myself ]
      set heading 90
      if (future-follower = nobody) or (abs([ future-acceleration myself ] of future-follower) < [ abs(max-deceleration) ] of future-follower)[

        ; Use lane-change-utility-force instead of lane-change-utility
        if not any? other turtles with [ ycor = lane and abs(xcor - [xcor] of myself) < 1 ] and lane-change-utility-force lane > currbest [
          set currbest lane-change-utility-force lane
          set target-lane lane
        ]

      ]
    ]
  ]
end

; ============= END-MOBIL - modified utility =============


; ============= BASELINE-1 COMMAND =============

to move-forward-baseline1
  foreach sort-on [(- xcor)] cars [ [vehicle] ->                ; Sort turtles based on xcor in decending order
    ask vehicle [
      if not crashed? and not blockade? [                       ; Do only if turtle has not crashed
        crash-check
        calculate-speeds                                        ; IDM
        check-lane-change                                       ; MOBIL
        if ycor != target-lane [ move-to-target-lane ]
      ]
      set heading 90
      fd speed
      set xcor precision xcor 6
    ]
  ]
end

; ============= END BASELINE-1 COMMAND =============


; ============= BASELINE-2 COMMAND =============

to move-forward-baseline2
  let block one-of blockades
  foreach sort-on [(- xcor)] cars [ [vehicle] ->                        ; Sort turtles based on xcor in decending order
    ask vehicle [
      if not crashed?[                                                  ; Do only if turtle has not crashed
        crash-check
        calculate-speeds                                                ; IDM
        ifelse xcor < red-zone-start-xcor or xcor > red-zone-end-xcor [
          check-lane-change                                             ; MOBIL
        ][
          check-lane-change-force                                       ; MOBIL - modified utility
        ]
        if ycor != target-lane [ move-to-target-lane ]
      ]
      set heading 90
      fd speed
      set xcor precision xcor 6
    ]
  ]
end

; ============= END BASELINE-2 COMMAND =============


; ============= STRATEGY COMMANDS =============

to move-forward [ strategy ]
  let block one-of blockades
  foreach sort-on [(- xcor)] cars [ [vehicle] ->                            ; Sort turtles based on xcor in decending order
    ask vehicle [
      crash-check
      if not crashed? and not blockade? [                                   ; Do only if turtle has not crashed
        ifelse xcor > red-zone-start-xcor and xcor < red-zone-end-xcor[
          ifelse ycor = [ycor] of block[
            blocked-lane-move strategy vehicle                              ; Cars in the red-zone and in the blocked lane
          ][
            unblocked-lane-move strategy vehicle                            ; Cars in the red-zone and in the unblocked lane
          ]
        ][
          calculate-speeds                                                  ; For cars not in the red-zone, follow IDM and MOBIL
          check-lane-change
        ]
      ]
      if ycor != target-lane [ move-to-target-lane ]
      set heading 90
      fd speed
      set xcor precision xcor 6
    ]
  ]
end

to blocked-lane-move [ strategy vehicle ]
  let block one-of blockades
  set heading -90
  let slow-down-car min-one-of ( other turtles in-cone world-width 180 with [ ycor != [ycor] of block ] ) [ distance myself ]
  set heading 90
  let speed-up-car min-one-of ( other turtles in-cone world-width 180 with [ ycor != [ycor] of block ] ) [ distance myself ]
  let new-speed ifelse-value (speed-up-car != nobody)[([speed] of speed-up-car) / 1.04][cruising-speed]
  if slow-down-car != nobody and [who-to-allow] of slow-down-car != vehicle [
    ask slow-down-car [
      set who-to-allow vehicle
      set slow-down-reqs slow-down-reqs + 1
      let old-slow-down-car min-one-of ( other turtles in-cone world-width 180 with [ ycor != [ycor] of block and who-to-allow = vehicle ] ) [ distance myself ]
      if old-slow-down-car != nobody [
        ask old-slow-down-car [
          set who-to-allow nobody
        ]
      ]
    ]
  ]
  (ifelse strategy = "STRATEGY-1"      [ decelerate-to new-speed ]
          strategy = "STRATEGY-2"      [ comf-decelerate-to new-speed ]
  )
  check-lane-change-force
end

to unblocked-lane-move [ strategy vehicle ]
  let block one-of blockades
  let faster-cars other turtles in-cone world-width 180 with [ ycor != [ycor] of block ]
  ifelse who-to-allow != nobody and [ycor] of who-to-allow != ycor and not member? who-to-allow faster-cars [
    (ifelse strategy = "STRATEGY-1"      [ decelerate-to max( list (([speed] of who-to-allow) / 1.04) 0.01) ]
            strategy = "STRATEGY-2"      [ comf-decelerate-to max( list (([speed] of who-to-allow) / 1.04) 0.01) ]
    )
  ][
    set who-to-allow nobody
    calculate-speeds
  ]
end

; ============= END STRATEGY COMMANDS =============


; ============= AUXILIARY COMMANDS =============

to check-for-halted-cars
  foreach sort-on [(- xcor)] cars [ [vehicle] ->
    ask vehicle [
      if speed = 0 [ set total-time-halted total-time-halted + 1 ]
    ]
  ]
end

; Reduce and maintain speeds without exceeding max-deceleration
to decelerate-to [req-speed]
  ifelse req-speed > (speed + acceleration) [ calculate-speeds ][
    set speed max(list req-speed (speed - max-deceleration))
  ]
end

to comf-decelerate-to [req-speed]
  ifelse req-speed > (speed + acceleration) or (-1 * comfortable-deceleration) > acceleration [ calculate-speeds ][
    set speed max(list req-speed (speed - comfortable-deceleration))
  ]
end

to move-to-target-lane
  set heading ifelse-value target-lane = ycor [ 90 ] [
    ifelse-value target-lane < ycor [ 180 ][ 0 ]
  ]
  let blocking-vehicles other turtles in-cone (1 + abs (ycor - target-lane)) 180 with [ x-distance <= 1 ]
  let blocking-vehicle min-one-of blocking-vehicles [ distance myself ]
  if blocking-vehicle = nobody [
    forward 2
    set ycor precision ycor 2
  ]
  set heading 90
end

to crash-check
  if any? other turtles in-cone 1 180 with [ y-distance < 1 and x-distance < 1 ][
    crash
    ask other turtles in-cone 1 180 with [ y-distance < 1 and x-distance < 1 ][ crash ]
    set crash-count (crash-count + 1)
  ]
end

to crash
  set color yellow
  set shape "fire"
  set crashed? true
  set crash-tick ticks
end

to-report x-distance
  report distancexy [ xcor ] of myself ycor
end

to-report y-distance
  report distancexy xcor [ ycor ] of myself
end

to-report vehicle-color
  report one-of [ blue cyan sky ] + 1.5 + random-float 1.0
end

to-report number-of-lanes
  report 2
end

to draw-red-zone
  ask patches with [ abs pycor <= number-of-lanes ] [ set pcolor grey - 2.5 + random-float 0.25 ]
  if any-lane-blocked? [
    ask patches with [ abs pycor <= number-of-lanes and pxcor > red-zone-start-xcor and pxcor < red-zone-end-xcor ] [ set pcolor red - 2.5 + random-float 0.25 ]
  ]
end

to toggle-blockade
  ifelse count turtles with [blockade? = true] = 0 [
    spawn-blockade one-of lanes
    set any-lane-blocked? true
  ][
    ask turtles with [blockade? = true] [ die ]
    set any-lane-blocked? false
  ]
  draw-red-zone
end

to crash-slow-down
  set speed max(list 0 (speed - 1))
end

; ============= END AUXILIARY COMMANDS =============


to setup
  clear-all
  reset-ticks
  set total-travel-time 0
  set acceleration-exponent 4
  set threshold 0
  set cruising-speed 0.625
  set blockade-xcor (((- world-width + 1)/ 2) + (1250 / 5))
  set any-lane-blocked? true
  set throughput 0
  set red-zone-start-xcor blockade-xcor - (1000 / 5)
  set red-zone-end-xcor blockade-xcor
  set spawn-count 0
  set non-realistic-brake-count 0
  set total-time-halted 0
  set cumulative-cars-in-world-count 0
  draw-road
  spawn-cars
  toggle-blockade
end

to go

  spawn-cars

  ; Kill turtles that are alive for more than 3 ticks after crashing
  ask cars with [ crashed? and ticks - crash-tick > 3][
    die
  ]

  ; Update car positions based on scenario
  (ifelse scenario = "Baseline-1"      [ move-forward-baseline1 ]
          scenario = "Baseline-2"      [ move-forward-baseline2 ]
          scenario = "Strategy-1"      [ move-forward "STRATEGY-1" ]
          scenario = "Strategy-2"      [ move-forward "STRATEGY-2" ]
  )

  ; Update total-time-halted
  check-for-halted-cars

  ; Kill turtles that reached the end of the world
  ask turtles with [xcor > ((world-width - 2) / 2)][
    set throughput throughput + 1
    set total-travel-time total-travel-time + ( ticks - entry-tick )
    die
  ]

  ; Update cumulative-cars-in-world-count
  set cumulative-cars-in-world-count cumulative-cars-in-world-count + count turtles with [blockade? = false]

  tick
end


; Copyright 2021 Hani Mohammed.
; See Info tab for full copyright and license.
@#$#@#$#@
GRAPHICS-WINDOW
225
10
2641
283
-1
-1
8.0
1
10
1
1
1
0
0
0
1
-150
150
-16
16
1
1
1
ticks
30.0

BUTTON
10
10
75
45
NIL
setup
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
150
10
215
45
go
go
T
1
T
OBSERVER
NIL
NIL
NIL
NIL
0

BUTTON
80
10
145
45
go once
go
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
0

MONITOR
395
425
510
470
Mean speed
(mean [speed] of turtles) * 10 * 5
2
1
11

MONITOR
10
335
205
380
Number of crashes
crash-count
2
1
11

SLIDER
10
125
220
158
politeness-factor-mean
politeness-factor-mean
0
1
0.4
0.001
1
NIL
HORIZONTAL

SLIDER
10
155
220
188
politeness-factor-deviation
politeness-factor-deviation
0
1
0.0
0.001
1
NIL
HORIZONTAL

BUTTON
10
50
215
83
NIL
toggle-blockade\n
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

MONITOR
280
335
395
380
Throughput
precision (throughput / (ticks / 600)) 3
3
1
11

CHOOSER
10
235
220
280
scenario
scenario
"Baseline-1" "Baseline-2" "Strategy-1" "Strategy-2"
0

PLOT
525
335
735
470
Mean Travel Time
ticks
NIL
0.0
1.0
0.0
1.0
true
false
"" ""
PENS
"default" 1.0 0 -16777216 true "" "ifelse spawn-count != 0 [plot ((total-travel-time / spawn-count) / 600)][plot 0]"

MONITOR
395
380
510
425
Spawn Rate
spawn-count / (ticks / 600)
3
1
11

MONITOR
10
380
205
425
Non realistic braking occurrences 
non-realistic-brake-count
17
1
11

MONITOR
395
335
510
380
Mean Halt Time
precision ((total-time-halted / 600) / spawn-count) 3
3
1
11

MONITOR
280
380
395
425
Mean Car Density
cumulative-cars-in-world-count / ticks
3
1
11

MONITOR
280
425
395
470
Mean Travel Time
precision ((total-travel-time / 600) / spawn-count) 3
17
1
11

TEXTBOX
400
305
700
341
QUALITY OF TRAFFIC FLOW
16
0.0
1

TEXTBOX
55
305
205
323
VERIFICATION
16
0.0
1

@#$#@#$#@
# Co-operative Autonomy

## WHAT IS IT?

This model presents an autonomous, two-lane driving environment with a single lane-closure that can be toggled. The four driving scenarios - two baseline cases (based on the real-world) and two experimental setups - are as follows:

- <b>Baseline-1</b> is where cars are not informed of the lane closure.
- <b>Baseline-2</b> is where a Red Zone is marked wherein cars are informed of the lane closure ahead.
- <b>Strategy-1</b> is where cars use a co-operative driving strategy - FAS. <sup>[1]</sup>
- <b>Strategy-2</b> is a variant of Strategy-1 and uses comfortable deceleration values instead of the vehicle's limit.

To simulate autonomous car-following behavior, we use the widely accepted Intelligent Driver Model (IDM) by Treiber (2000). In order to allow cars modeled with IDM to make lane changes in our two-lane freeway, we use a lane change model - Minimizing Overall Braking Induced by Lane changes (MOBIL) - by Kesting et al. (2007).

The results folder contains a spreadsheet where we compare each driving scenario with the others. Every metric is also tested for statistical significance. For more details, refer [1].

## HOW TO USE IT

Click on the SETUP button to set up the world. Click on GO to start the simulated traffic flow. The GO ONCE button simulates the traffic flow for one tick of the clock (100ms).

The TOGGLE-BLOCKADE button toggles the presence of a blockade (and a Red Zone).

The POLITENESS-FACTOR-MEAN and POLITENESS-FACTOR-DEVIATION sliders control the Normal (Gaussian) distribution of politeness factor *p* values of the cars on the road. The politeness factor *p* determines the degree to which a car’s neighbors influence its lane-changing decisions. <sup>[3]</sup>

The SCENARIO dropdown lets you choose the driving scenario to be simulated.

You may wish to slow down the model with the speed slider to watch the driving and lane-changing decisions of certain cars more closely.

The NUMBER OF CRASHES and NON-REALISTIC BRAKING OCCURRENCES monitors help validate the simulation. Ideally, a driving strategy for autonomous cars should not give rise to crashes. Similarly, no car should be asked to decelerate over its braking limit.

The MEAN SPEED monitor shows the average speed in *m/s* for the cars currently in the world.

The quality of traffic flow can be measured with the following monitors:

- **Throughput *(cars/min)*:** The number of cars that exit the world per tick.

- **Mean Halt Time *(min)*:** The average time that a car remains stationary for while transiting through the world.

- **Mean Car Density *(cars/15000 m<sup>2</sup>)*:** The average number of cars present in the world, which has an area of 15000 m<sup>2</sup>, at any given instance.

- **Spawn Rate *(cars/min)*:** The number of cars that enter the world per minute.

- **Mean Travel Time *(min)*:** The average time that a car takes to transit through the entire length of the world.

The MEAN TRAVEL TIME plot helps visualize the evolution of the quality of traffic flow over time.


## THINGS TO NOTICE

In the baseline cases, cars start queueing up behind the blockade right from the start of the simulation. Changing the politeness factor mean to a value closer to 0 will cause a few cars from the blocked lane to cut into the unblocked lane but soon afterwards, the result remains the same - long queues are formed behind the blockade. Some cars remain idle for long intervals while cars that were spawned after them pass through. This seems to be incredibly unfair but at the same time, is quite expected on a freeway where cars are cruising at high speeds and cutting into another lane is not a vaible option.

In the experimental setups, cars co-operate using FAS and pass through the blockade without remaining idle. Only when the amount of traffic is too high do cars start piling up. But even then, cars do not remain idle for more than a few seconds.

Lane changes in FAS often have a cascading effect, where one lane change in the system induces lane changes in the cars behind it. When a lane change happens in the front, the new follower is forced to brake slightly. Consequently, for the cars behind that are in the blocked lane, MOBIL decides that the overall braking is minimized by a lane change rather than staying on the blocked lane which has a negative incentive.

Interestingly, Strategy-1 and Strategy-2 show virtually no difference. Since cars have the same characteristics, the difference in speeds between cars is usually minimal. Hence, in order to match their predecessor's speed as FAS mandates, cars do not have to brake more than their capacity and a comfortable deceleration is sufficient.

## THINGS TO TRY

How would heterogeneous traffic affect the strategies? Will including vehicles like trucks, which have much different characteristics from cars, affect the performance of Strategy-1 and Strategy-2?

Will increasing the number of lanes require modifications to the strategies? What about multiple lane closures in three-lane freeways? These could also be on the extreme lanes or on adjacent lanes. What if blockades are staggered?

Can High Priority Vehicles like ambulances be accomodated? What about lane-change trajectories? Can vehicular platoons help improve the strategies?

## EXTENDING THE MODEL

Try adding your own strategies to the model. Define your own `move-forward-custom` function and run them in the `go` routine.

Add your own vehicle types by creating new breeds of turtles. Try and mix up the spawn command to maintain a given propotion of cars to this new vehicle type.

The way this model is written makes it easy to add more lanes. Look for the `number-of-lanes` reporter in the code and play around with it.

## NETLOGO FEATURES

Behaviour space experiments help run the simulation for extended amounts of time while also running them with different parameters. This is really important if you want to gather statistically significant results to back up claims made on different driving scenarios.

Each turtle has a shape, unlike in some other models. NetLogo uses `set shape` to alter the shapes of turtles. You can, using the shapes editor in the Tools menu, create your own turtle shapes or modify existing ones. Then you can modify the code to use your own shapes.

## RELATED MODELS

- "Traffic Basic": a simple model of the movement of cars on a highway.

- "Traffic Basic Utility": a version of "Traffic Basic" including a utility function for the cars.

- "Traffic Basic Adaptive": a version of "Traffic Basic" where cars adapt their acceleration to try and maintain a smooth flow of traffic.

- "Traffic Basic Adaptive Individuals": a version of "Traffic Basic Adaptive" where each car adapts individually, instead of all cars adapting in unison.

## REFERENCES

[1] Mohammed, Hani, Venkat Himavanth Reddy, and Subu Kandaswamy. "A Decentralized Strategy for Cooperative Driving among Autonomous Cars at Lane Closures." 2020 IEEE 32nd International Conference on Tools with Artificial Intelligence (ICTAI). IEEE, 2020.

[2] Treiber, Martin, Ansgar Hennecke, and Dirk Helbing.” Congested traffic
states in empirical observations and microscopic simulations.” Physical
review E 62.2 (2000): 1805.

[3] Kesting, Arne, Martin Treiber, and Dirk Helbing.” General lane-changing model MOBIL for car-following models.” Transportation Research Record 1999.1 (2007): 86-94.

## HOW TO CITE

If you mention this model in a publication, we ask that you include the citation below:

* Mohammed, Hani, Venkat Himavanth Reddy, and Subu Kandaswamy. "A Decentralized Strategy for Cooperative Driving among Autonomous Cars at Lane Closures." 2020 IEEE 32nd International Conference on Tools with Artificial Intelligence (ICTAI). IEEE, 2020.

## COPYRIGHT AND LICENSE

Copyright 2021 Hani Mohammed.

![CC BY-NC-SA 3.0](http://ccl.northwestern.edu/images/creativecommons/byncsa.png)

This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 4.0 License.  To view a copy of this license, visit https://creativecommons.org/licenses/by-nc-sa/4.0/ or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.

Commercial licenses are also available. To inquire about commercial licenses, please contact Subu Kandaswamy at subu.kandaswamy@gmail.com.
@#$#@#$#@
default
true
0
Polygon -7500403 true true 150 5 40 250 150 205 260 250

airplane
true
0
Polygon -7500403 true true 150 0 135 15 120 60 120 105 15 165 15 195 120 180 135 240 105 270 120 285 150 270 180 285 210 270 165 240 180 180 285 195 285 165 180 105 180 60 165 15

arrow
true
0
Polygon -7500403 true true 150 0 0 150 105 150 105 293 195 293 195 150 300 150

bikes-top
true
0
Polygon -7500403 true true 135 180 120 135 120 105 135 90 165 90 180 105 180 135 165 180 135 180
Polygon -16777216 true false 165 285 135 285 120 195 135 165 165 165 180 195 165 285
Rectangle -16777216 true false 142 25 158 63
Polygon -7500403 true true 120 90 180 90 165 60 135 60 120 90
Polygon -7500403 true true 136 91 107 106 106 102 137 82 138 90
Polygon -7500403 true true 163 82 194 103 191 107 159 88 163 83
Polygon -16777216 true false 193 106 214 125 214 113 194 102
Polygon -16777216 true false 107 106 86 125 86 113 106 102

box
false
0
Polygon -7500403 true true 150 285 285 225 285 75 150 135
Polygon -7500403 true true 150 135 15 75 150 15 285 75
Polygon -7500403 true true 15 75 15 225 150 285 150 135
Line -16777216 false 150 285 150 135
Line -16777216 false 150 135 15 75
Line -16777216 false 150 135 285 75

bug
true
0
Circle -7500403 true true 96 182 108
Circle -7500403 true true 110 127 80
Circle -7500403 true true 110 75 80
Line -7500403 true 150 100 80 30
Line -7500403 true 150 100 220 30

buses-top
true
0
Polygon -7500403 true true 150 15 120 15 90 30 90 45 90 225 90 270 105 285 150 285 195 285 210 270 210 225 210 45 210 30 180 15
Polygon -16777216 true false 106 58 194 58 204 44 182 36 153 32 120 36 98 44
Polygon -1 true false 205 29 180 30 182 16
Polygon -1 true false 95 29 120 30 118 16

butterfly
true
0
Polygon -7500403 true true 150 165 209 199 225 225 225 255 195 270 165 255 150 240
Polygon -7500403 true true 150 165 89 198 75 225 75 255 105 270 135 255 150 240
Polygon -7500403 true true 139 148 100 105 55 90 25 90 10 105 10 135 25 180 40 195 85 194 139 163
Polygon -7500403 true true 162 150 200 105 245 90 275 90 290 105 290 135 275 180 260 195 215 195 162 165
Polygon -16777216 true false 150 255 135 225 120 150 135 120 150 105 165 120 180 150 165 225
Circle -16777216 true false 135 90 30
Line -16777216 false 150 105 195 60
Line -16777216 false 150 105 105 60

car
false
0
Polygon -7500403 true true 300 180 279 164 261 144 240 135 226 132 213 106 203 84 185 63 159 50 135 50 75 60 0 150 0 165 0 225 300 225 300 180
Circle -16777216 true false 180 180 90
Circle -16777216 true false 30 180 90
Polygon -16777216 true false 162 80 132 78 134 135 209 135 194 105 189 96 180 89
Circle -7500403 true true 47 195 58
Circle -7500403 true true 195 195 58

cars-top
true
0
Polygon -7500403 true true 151 8 119 10 98 25 86 48 82 225 90 270 105 289 150 294 195 291 210 270 219 225 214 47 201 24 181 11
Polygon -16777216 true false 210 195 195 210 195 135 210 105
Polygon -16777216 true false 105 255 120 270 180 270 195 255 195 225 105 225
Polygon -16777216 true false 90 195 105 210 105 135 90 105
Polygon -1 true false 205 29 180 30 181 11
Line -7500403 true 210 165 195 165
Line -7500403 true 90 165 105 165
Polygon -16777216 true false 121 135 180 134 204 97 182 89 153 85 120 89 98 97
Line -16777216 false 210 90 195 30
Line -16777216 false 90 90 105 30
Polygon -1 true false 95 29 120 30 119 11

circle
false
0
Circle -7500403 true true 0 0 300

circle 2
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240

cow
false
0
Polygon -7500403 true true 200 193 197 249 179 249 177 196 166 187 140 189 93 191 78 179 72 211 49 209 48 181 37 149 25 120 25 89 45 72 103 84 179 75 198 76 252 64 272 81 293 103 285 121 255 121 242 118 224 167
Polygon -7500403 true true 73 210 86 251 62 249 48 208
Polygon -7500403 true true 25 114 16 195 9 204 23 213 25 200 39 123

cylinder
false
0
Circle -7500403 true true 0 0 300

dot
false
0
Circle -7500403 true true 90 90 120

face happy
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 255 90 239 62 213 47 191 67 179 90 203 109 218 150 225 192 218 210 203 227 181 251 194 236 217 212 240

face neutral
false
0
Circle -7500403 true true 8 7 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Rectangle -16777216 true false 60 195 240 225

face sad
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 168 90 184 62 210 47 232 67 244 90 220 109 205 150 198 192 205 210 220 227 242 251 229 236 206 212 183

fire
false
0
Polygon -7500403 true true 151 286 134 282 103 282 59 248 40 210 32 157 37 108 68 146 71 109 83 72 111 27 127 55 148 11 167 41 180 112 195 57 217 91 226 126 227 203 256 156 256 201 238 263 213 278 183 281
Polygon -955883 true false 126 284 91 251 85 212 91 168 103 132 118 153 125 181 135 141 151 96 185 161 195 203 193 253 164 286
Polygon -2674135 true false 155 284 172 268 172 243 162 224 148 201 130 233 131 260 135 282

fish
false
0
Polygon -1 true false 44 131 21 87 15 86 0 120 15 150 0 180 13 214 20 212 45 166
Polygon -1 true false 135 195 119 235 95 218 76 210 46 204 60 165
Polygon -1 true false 75 45 83 77 71 103 86 114 166 78 135 60
Polygon -7500403 true true 30 136 151 77 226 81 280 119 292 146 292 160 287 170 270 195 195 210 151 212 30 166
Circle -16777216 true false 215 106 30

flag
false
0
Rectangle -7500403 true true 60 15 75 300
Polygon -7500403 true true 90 150 270 90 90 30
Line -7500403 true 75 135 90 135
Line -7500403 true 75 45 90 45

flower
false
0
Polygon -10899396 true false 135 120 165 165 180 210 180 240 150 300 165 300 195 240 195 195 165 135
Circle -7500403 true true 85 132 38
Circle -7500403 true true 130 147 38
Circle -7500403 true true 192 85 38
Circle -7500403 true true 85 40 38
Circle -7500403 true true 177 40 38
Circle -7500403 true true 177 132 38
Circle -7500403 true true 70 85 38
Circle -7500403 true true 130 25 38
Circle -7500403 true true 96 51 108
Circle -16777216 true false 113 68 74
Polygon -10899396 true false 189 233 219 188 249 173 279 188 234 218
Polygon -10899396 true false 180 255 150 210 105 210 75 240 135 240

house
false
0
Rectangle -7500403 true true 45 120 255 285
Rectangle -16777216 true false 120 210 180 285
Polygon -7500403 true true 15 120 150 15 285 120
Line -16777216 false 30 120 270 120

leaf
false
0
Polygon -7500403 true true 150 210 135 195 120 210 60 210 30 195 60 180 60 165 15 135 30 120 15 105 40 104 45 90 60 90 90 105 105 120 120 120 105 60 120 60 135 30 150 15 165 30 180 60 195 60 180 120 195 120 210 105 240 90 255 90 263 104 285 105 270 120 285 135 240 165 240 180 270 195 240 210 180 210 165 195
Polygon -7500403 true true 135 195 135 240 120 255 105 255 105 285 135 285 165 240 165 195

line
true
0
Line -7500403 true 150 0 150 300

line half
true
0
Line -7500403 true 150 0 150 150

pentagon
false
0
Polygon -7500403 true true 150 15 15 120 60 285 240 285 285 120

person
false
0
Circle -7500403 true true 110 5 80
Polygon -7500403 true true 105 90 120 195 90 285 105 300 135 300 150 225 165 300 195 300 210 285 180 195 195 90
Rectangle -7500403 true true 127 79 172 94
Polygon -7500403 true true 195 90 240 150 225 180 165 105
Polygon -7500403 true true 105 90 60 150 75 180 135 105

plant
false
0
Rectangle -7500403 true true 135 90 165 300
Polygon -7500403 true true 135 255 90 210 45 195 75 255 135 285
Polygon -7500403 true true 165 255 210 210 255 195 225 255 165 285
Polygon -7500403 true true 135 180 90 135 45 120 75 180 135 210
Polygon -7500403 true true 165 180 165 210 225 180 255 120 210 135
Polygon -7500403 true true 135 105 90 60 45 45 75 105 135 135
Polygon -7500403 true true 165 105 165 135 225 105 255 45 210 60
Polygon -7500403 true true 135 90 120 45 150 15 180 45 165 90

rickshaws-top
true
4
Rectangle -16777216 true false 60 210 75 270
Rectangle -16777216 true false 225 210 240 270
Polygon -1184463 true true 240 270 225 90 210 60 180 45 120 45 90 60 75 90 60 270 225 270 225 270
Rectangle -16777216 true false 144 6 157 31
Polygon -7500403 true false 120 45 135 30 165 30 180 45 120 45

square
false
0
Rectangle -7500403 true true 30 30 270 270

square 2
false
0
Rectangle -7500403 true true 30 30 270 270
Rectangle -16777216 true false 60 60 240 240

star
false
0
Polygon -7500403 true true 151 1 185 108 298 108 207 175 242 282 151 216 59 282 94 175 3 108 116 108

target
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240
Circle -7500403 true true 60 60 180
Circle -16777216 true false 90 90 120
Circle -7500403 true true 120 120 60

tree
false
0
Circle -7500403 true true 118 3 94
Rectangle -6459832 true false 120 195 180 300
Circle -7500403 true true 65 21 108
Circle -7500403 true true 116 41 127
Circle -7500403 true true 45 90 120
Circle -7500403 true true 104 74 152

triangle
false
0
Polygon -7500403 true true 150 30 15 255 285 255

triangle 2
false
0
Polygon -7500403 true true 150 30 15 255 285 255
Polygon -16777216 true false 151 99 225 223 75 224

truck
false
0
Rectangle -7500403 true true 4 45 195 187
Polygon -7500403 true true 296 193 296 150 259 134 244 104 208 104 207 194
Rectangle -1 true false 195 60 195 105
Polygon -16777216 true false 238 112 252 141 219 141 218 112
Circle -16777216 true false 234 174 42
Rectangle -7500403 true true 181 185 214 194
Circle -16777216 true false 144 174 42
Circle -16777216 true false 24 174 42
Circle -7500403 false true 24 174 42
Circle -7500403 false true 144 174 42
Circle -7500403 false true 234 174 42

trucks-top
true
0
Polygon -7500403 true true 150 15 135 15 118 23 109 45 104 75 105 114 150 114
Polygon -7500403 true true 150 15 165 15 182 23 191 45 196 75 195 114 150 114
Rectangle -7500403 true true 90 120 210 285
Rectangle -7500403 true true 135 120 165 120
Rectangle -7500403 true true 135 105 165 135
Rectangle -16777216 true false 120 60 180 75
Polygon -16777216 true false 180 60 164 51 148 51 149 61
Polygon -16777216 true false 120 60 136 51 152 51 151 61
Polygon -1184463 true false 137 16 137 26 120 25
Polygon -1184463 true false 163 16 163 26 180 25
Line -16777216 false 193 136 194 275
Line -16777216 false 107 136 106 275
Line -16777216 false 177 139 177 279
Line -16777216 false 123 139 123 279
Line -16777216 false 158 143 157 279
Line -16777216 false 142 143 143 279

turtle
true
0
Polygon -10899396 true false 215 204 240 233 246 254 228 266 215 252 193 210
Polygon -10899396 true false 195 90 225 75 245 75 260 89 269 108 261 124 240 105 225 105 210 105
Polygon -10899396 true false 105 90 75 75 55 75 40 89 31 108 39 124 60 105 75 105 90 105
Polygon -10899396 true false 132 85 134 64 107 51 108 17 150 2 192 18 192 52 169 65 172 87
Polygon -10899396 true false 85 204 60 233 54 254 72 266 85 252 107 210
Polygon -7500403 true true 119 75 179 75 209 101 224 135 220 225 175 261 128 261 81 224 74 135 88 99

wheel
false
0
Circle -7500403 true true 3 3 294
Circle -16777216 true false 30 30 240
Line -7500403 true 150 285 150 15
Line -7500403 true 15 150 285 150
Circle -7500403 true true 120 120 60
Line -7500403 true 216 40 79 269
Line -7500403 true 40 84 269 221
Line -7500403 true 40 216 269 79
Line -7500403 true 84 40 221 269

x
false
0
Polygon -7500403 true true 270 75 225 30 30 225 75 270
Polygon -7500403 true true 30 75 75 30 270 225 225 270
@#$#@#$#@
NetLogo 6.1.0
@#$#@#$#@
@#$#@#$#@
@#$#@#$#@
<experiments>
  <experiment name="Compare-all" repetitions="20" runMetricsEveryStep="false">
    <setup>setup</setup>
    <go>go</go>
    <timeLimit steps="36000"/>
    <metric>precision (throughput / (ticks / 600)) 3</metric>
    <metric>precision ((total-travel-time / spawn-count) / 600) 3</metric>
    <metric>precision ((total-time-halted / spawn-count) / 600) 3</metric>
    <metric>precision (cumulative-cars-in-world-count / ticks) 3</metric>
    <metric>spawn-count / (ticks / 600)</metric>
    <metric>crash-count</metric>
    <metric>non-realistic-brake-count</metric>
    <enumeratedValueSet variable="politeness-factor-deviation">
      <value value="0"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="scenario">
      <value value="&quot;Baseline-1&quot;"/>
      <value value="&quot;Baseline-2&quot;"/>
      <value value="&quot;Strategy-1&quot;"/>
      <value value="&quot;Strategy-2&quot;"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="politeness-factor-mean">
      <value value="0.2"/>
      <value value="0.3"/>
      <value value="0.4"/>
      <value value="0.5"/>
      <value value="0.6"/>
    </enumeratedValueSet>
  </experiment>
</experiments>
@#$#@#$#@
@#$#@#$#@
default
0.0
-0.2 0 0.0 1.0
0.0 1 1.0 0.0
0.2 0 0.0 1.0
link direction
true
0
Line -7500403 true 150 150 90 180
Line -7500403 true 150 150 210 180
@#$#@#$#@
1
@#$#@#$#@
